/*
 *      "History" SQLite extension
 *
 *      Alexey Pechnikov <pechnikov@mobigroup.ru>
 *      The code is public domain.
 */

/* 
Compile as
     gcc -shared history.c -o libsqlitehistory.so
*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_HISTORY)

#include <assert.h>
#include <stdlib.h>

#ifndef NDEBUG
    #include <string.h>
#endif 

#ifndef SQLITE_CORE
    #include <sqlite3ext.h>
    SQLITE_EXTENSION_INIT1
#else
    #include "sqlite3.h"
#endif

static void history(sqlite3_context *context, int argc,
    sqlite3_value **argv)
{
    int rc, pk;
    char *column = "rowid", *table;
    char *sql, *name, *names = NULL, *_names = NULL;
    sqlite3_stmt *stmt;
    sqlite3 *db;

    if (argc != 2) {
        sqlite3_result_error(context, "Function must get two arguments. First argument is table name, "
            "and the second is key column name", -1);
        return;
    }

    if (sqlite3_value_type(argv[0]) != SQLITE_TEXT) {
        sqlite3_result_error(context, "The first argument is table name and must be a text string", -1);
        return;
    }
    table = (char*)sqlite3_value_text(argv[0]);

    if (sqlite3_value_type(argv[1]) != SQLITE_TEXT) {
        sqlite3_result_error(context, "The second argument is the key column name and must be a text string", -1);
        return;
    }
    column  = (char*)sqlite3_value_text(argv[1]);

    db = sqlite3_context_db_handle(context);

    rc = sqlite3_exec(db, "SAVEPOINT history;", NULL, NULL, NULL);
    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    sql = sqlite3_mprintf(
        // add history table
        "CREATE TABLE IF NOT EXISTS %s_history AS "
              "SELECT *, "
              "CAST(0 AS REAL) as _date, "
              "CAST('' AS TEXT) as _action, "
              "CAST('' AS TEXT) as _user "
            "FROM %s LIMIT 0;" // this work, really work :-)
        "INSERT INTO %s_history (%s, _date, _action, _user) SELECT %s, julianday('now'), '', user() FROM %s where (select count(*) from %s_history)=0;"
        "CREATE INDEX IF NOT EXISTS %s_history_complex_idx on %s_history(%s, _date);"
        // add triggers to source table
        "CREATE TRIGGER IF NOT EXISTS %s_history_i AFTER INSERT ON %s BEGIN "
            "INSERT INTO %s_history (%s, _date, _action, _user) "
               "VALUES (NEW.%s, julianday('now'), 'I', user());"
        "END;"
        "CREATE TRIGGER IF NOT EXISTS %s_history_u BEFORE UPDATE ON %s BEGIN "
            "SELECT RAISE (ABORT,'The key field \"%s\" can not be modified') WHERE NEW.%s != OLD.%s;"
            "INSERT INTO %s_history SELECT *, julianday('now'), 'U', user() FROM %s WHERE rowid=OLD.rowid;"
        "END;"
        "CREATE TRIGGER IF NOT EXISTS %s_history_d BEFORE DELETE ON %s BEGIN "
            "INSERT INTO %s_history SELECT *, julianday('now'), 'D', user() FROM %s WHERE rowid=OLD.rowid;"
        "END;",
        // history table
        table, table,
        // initial import
        table, column, column, table, table,
        // indices
        table, table, column,
        // insert trigger
        table, table, table, column, column,
        // update trigger
        table, table, column, column, column, table, table,
        // delete trigger
        table, table, table, table
        );
//sqlite3_result_text( context, (char*)sql, -1, SQLITE_TRANSIENT);
//return;
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        sqlite3_exec(db, "ROLLBACK TO SAVEPOINT history;", NULL, NULL, NULL);
    }
    sqlite3_exec(db, "RELEASE SAVEPOINT history;", NULL, NULL, NULL);

    sqlite3_result_null(context);
}

static void unhistory(sqlite3_context *context, int argc,
            sqlite3_value **argv)
{
    int rc;
    char *table;
    char *sql = NULL;
    sqlite3_stmt *stmt;
    sqlite3 *db;

    if (argc != 1) {
        sqlite3_result_error(context, "Function must get single argument. The argument is table name.", -1);
        return;
    }
    if ((sqlite3_value_type(argv[0]) != SQLITE_TEXT) && (sqlite3_value_type(argv[0]) != SQLITE_INTEGER)) {
        sqlite3_result_error(context, "Table name must be a text string", -1);
        return;
    }

    table = (char*)sqlite3_value_text(argv[0]);
    db = sqlite3_context_db_handle(context);

    rc = sqlite3_exec(db, "SAVEPOINT history;", NULL, NULL, NULL);
    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    sql = sqlite3_mprintf(
        // drop triggers
        "DROP TRIGGER IF EXISTS %s_history_i;"
        "DROP TRIGGER IF EXISTS %s_history_u;"
        "DROP TRIGGER IF EXISTS %s_history_d;",
        table, table, table
        );
//sqlite3_result_text( context, (char*)sql, -1, SQLITE_TRANSIENT);
//return;
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        sqlite3_exec(db, "ROLLBACK TO SAVEPOINT history;", NULL, NULL, NULL);
        return;
    }

    sqlite3_exec(db, "RELEASE SAVEPOINT history;", NULL, NULL, NULL);

    sqlite3_result_null(context);
}


#if !SQLITE_CORE
static
#endif
int sqlite3HistoryInit(sqlite3 *db)
{
    int rc, i;
    static const struct {
        char *zName;
        signed char nArg;
        int argType;           /* 1: 0, 2: 1, 3: 2,...  N:  N-1. */
        int eTextRep;          /* 1: UTF-16.  0: UTF-8 */
        void (*xFunc)(sqlite3_context*,int,sqlite3_value **);
    } aFuncs[] = {
        { "history",               2, 0, SQLITE_UTF8,    history        },
        { "unhistory",             1, 0, SQLITE_UTF8,    unhistory      },
    };

    for(i=0; i<sizeof(aFuncs)/sizeof(aFuncs[0]); i++){
        void *pArg;
        int argType = aFuncs[i].argType;
        pArg = (void*)(int)argType;
        rc = sqlite3_create_function(db, aFuncs[i].zName, aFuncs[i].nArg,
            aFuncs[i].eTextRep, pArg, aFuncs[i].xFunc, 0, 0);
        if (rc != SQLITE_OK) {
            break;
        }
    }

    return rc;
}

#if !SQLITE_CORE
int sqlite3_extension_init(sqlite3 *db, char **pzErrMsg,
                        const sqlite3_api_routines *pApi)
{
    SQLITE_EXTENSION_INIT2(pApi)
    return sqlite3HistoryInit(db);
}
#endif

#endif
