/*
The functions coded by Alexey Pechnikov (pechnikov@mobigroup.ru) and tested on linux only.
The code is public domain.

1) Compile with
Linux:
    gcc -fPIC -lm -shared key.c -o libsqlitekey.so

Use as:

    .load './libsqlitekey.so'
or
    SELECT load_extension('./libsqlitekey.so');
or
    SELECT load_extension('/usr/lib/sqlite3/libsqlitekey.so');


Usage:

Init operations:

    Create "keys" table
    => select key_install();

    Drop "keys" table
    => select key_uninstall(); -- Fix it: no work now!


Base operations:

    Check is param exists
    => select key_exists('common','nds%');
    0
    Get _non-existing_ value _without_ default_value
    => select key_get('common','nds%');

    Get _non-existing_ value _with_ default_value
    => select key_get('common','nds%',18);
    18

    Set value
    => select key_set('common','nds%',10);
    10

    Get _existing_ value
    => select key_get('common','nds%');
    10

    Add _existing_ key
    => select key_add('common','nds%',18);

    Replace _existing_ key
    => select key_replace('common','nds%',20);
    20

    Delete existing key
    => select key_delete('common','nds%');
    1

    Delete non-existing key
    => select key_delete('common','nds%');

    Delete all keys in unit
    => select key_delete('common');

    Replace _non-existing_ key
    => select key_replace('common','nds%',20);

    Add _non-existing_ key
    => select key_add('common','nds%',18);
    18


Numeric operations:

    Increment _non-existing_ value
    => select key_increment('common','num',10);
    10

    Increment _existing_ value
    => select key_increment('common','num',10);
    20


String operations:

    Prepend value
    => select key_prepend('common','string','prepend-');
    prepend-

    Append value
    => select key_append('common','string','-append');
    prepend--append


Note:

    key_flush is synonym to key_delete

*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_KEY)

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#ifndef SQLITE_CORE
#include "sqlite3ext.h"
SQLITE_EXTENSION_INIT1
#else
#include "sqlite3.h"
#endif

#define KEY_SCHEMA_CREATE \
"CREATE TABLE IF NOT EXISTS _keys (  -- persistent parameters storage \n" \
"        unit text not null,\n"                                          \
"        param text not null,\n"                                         \
"        value text not null,\n"                                         \
"        UNIQUE (unit,param) on conflict replace\n"                      \
");"

//#define KEY_SCHEMA_DROP "DROP TABLE IF EXISTS _keys;"
#define KEY_SCHEMA_DROP "DELETE FROM _keys;"

#define KEY_DELETE_UNIT      "delete from _keys where unit=?"
#define KEY_DELETE           "delete from _keys where unit=? and param=?"
#define KEY_INSERT           "insert into _keys (unit,param,value) values (?,?,?)"
#define KEY_UPDATE           "update _keys set value=? where unit=? and param=?"
#define KEY_INCREMENT        "update _keys set value=value+? where unit=? and param=?"
#define KEY_APPEND           "update _keys set value=value || ? where unit=? and param=?"
#define KEY_PREPEND          "update _keys set value=? || value where unit=? and param=?"
#define KEY_COUNT            "select count(*) from _keys where unit=? and param=?"
#define KEY_SELECT           "select value from _keys where unit=? and param=?"

int key_count(sqlite3_context *context, unsigned char* unit, unsigned char* param, int* count){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */
    *count = 0;

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_COUNT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    if(SQLITE_ROW == rc) {
        *count = sqlite3_column_int(pStmt,0);
    }
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int key_insert(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_INSERT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, value, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int key_update(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_UPDATE, -1, &pStmt, 0);
    if( SQLITE_OK != rc){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int key_increment(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_INCREMENT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int key_append(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_APPEND, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}


int key_prepend(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_PREPEND, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int key_delete(sqlite3_context *context, unsigned char* unit, unsigned char* param){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_DELETE, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param,  -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}


int key_delete_unit(sqlite3_context *context, unsigned char* unit){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_DELETE_UNIT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

// returned value must be cleaned!
int key_get(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char** value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, KEY_SELECT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    if( SQLITE_ROW == rc ) {
        *value = (unsigned char*)sqlite3_malloc (strlen((unsigned char*)sqlite3_column_text(pStmt,0))+1);
        strcpy( *value, (unsigned char*)sqlite3_column_text(pStmt,0));
    }
    sqlite3_finalize(pStmt);
    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

// key_exists (unit,param)
static void key_existsFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int count;
    int rc;                     /* Result code */

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    sqlite3_result_int(context, count);
}

// key_get (unit,param,default_value)
static void key_getFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return default_value when defined 3-th argument
    if ( 0 == count  && 3 == argc ) {
        sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
        return;
    }
    // key not found
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }

    rc = key_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    sqlite3_result_text( context, value, -1, sqlite3_free);

}

// store this data
// key_set (unit,param,value)
static void key_setFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = key_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = key_update(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // returns value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// increment this data
// key_increment (unit,param,value)
static void key_incrementFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = key_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = key_increment(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = key_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// append this data
// key_append (unit,param,value)
static void key_appendFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value = 0;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = key_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = key_append(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = key_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// prepend this data
// key_prepend (unit,param,value)
static void key_prependFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value = 0;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = key_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = key_prepend(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = key_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// store this data, but only if the server *doesn't* already hold data for this key
// key_add (unit,param,value)
static void key_addFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server does already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (count > 0) {
        sqlite3_result_null(context);
        return;
    }
    rc = key_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return added value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// store this data, but only if the server *does* already hold data for this key
// key_replace (unit,param,value)
static void key_replaceFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }
    rc = key_update(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return replaced value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// delete this data
// key_delete (unit,param)
static void key_deleteFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't hold this key
    rc = key_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }
    rc = key_delete(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return 1
    sqlite3_result_int(context, 1);
}

// delete this unit
// key_delete (unit)
static void key_delete_unitFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = key_delete_unit(context, (unsigned char*) sqlite3_value_text(argv[0]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

static void key_installFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_exec(db, KEY_SCHEMA_CREATE, 0, 0, 0);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

static void key_uninstallFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_exec(db, KEY_SCHEMA_DROP, 0, 0, 0);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

/* SQLite invokes this routine once when it loads the extension.
** Create new functions, collating sequences, and virtual table
** modules here.  This is usually the only exported symbol in
** the shared library.
*/

int sqlite3KeyInit(sqlite3 *db){
static const struct {
    char *zName;
    signed char nArg;
    int argType;           /* 1: 0, 2: 1, 3: 2,...  N:  N-1. */
    int eTextRep;          /* 1: UTF-16.  0: UTF-8 */
    void (*xFunc)(sqlite3_context*,int,sqlite3_value **);
} aFuncs[] = {
    { "key_install",             0, 0, SQLITE_UTF8,    key_installFunc },
    { "key_uninstall",           0, 0, SQLITE_UTF8,    key_uninstallFunc },
    { "key_exists",              2, 0, SQLITE_UTF8,    key_existsFunc },
    { "key_get",                 2, 0, SQLITE_UTF8,    key_getFunc },
    { "key_get",                 3, 0, SQLITE_UTF8,    key_getFunc },
    { "key_delete",              2, 0, SQLITE_UTF8,    key_deleteFunc },
    { "key_delete",              1, 0, SQLITE_UTF8,    key_delete_unitFunc },
    { "key_flush",               2, 0, SQLITE_UTF8,    key_deleteFunc },
    { "key_flush",               1, 0, SQLITE_UTF8,    key_delete_unitFunc },
    { "key_set",                 3, 0, SQLITE_UTF8,    key_setFunc },
    { "key_increment",           3, 0, SQLITE_UTF8,    key_incrementFunc },
    { "key_append",              3, 0, SQLITE_UTF8,    key_appendFunc },
    { "key_prepend",             3, 0, SQLITE_UTF8,    key_prependFunc },
    { "key_add",                 3, 0, SQLITE_UTF8,    key_addFunc },
    { "key_replace",             3, 0, SQLITE_UTF8,    key_replaceFunc },
};

int i;
for(i=0; i<sizeof(aFuncs)/sizeof(aFuncs[0]); i++){
    void *pArg;
    int argType = aFuncs[i].argType;
    pArg = (void*)(int)argType;
    sqlite3_create_function(db, aFuncs[i].zName, aFuncs[i].nArg,
        aFuncs[i].eTextRep, pArg, aFuncs[i].xFunc, 0, 0);
}

return 0;
}

#if !SQLITE_CORE
int sqlite3_extension_init(
sqlite3 *db,
char **pzErrMsg,
const sqlite3_api_routines *pApi
){
SQLITE_EXTENSION_INIT2(pApi)
return sqlite3KeyInit(db);
}
#endif

#endif
