/*
The functions coded by Alexey Pechnikov (pechnikov@mobigroup.ru) and tested on linux only.
The code is public domain.

1) Compile with
Linux:
    gcc -fPIC -lm -shared tempkey.c -o libsqlitetempkey.so

Use as:

    .load './libsqlitetempkey.so'
or
    SELECT load_extension('./libsqlitetempkey.so');
or
    SELECT load_extension('/usr/lib/sqlite3/libsqlitetempkey.so');


Usage:

Init operations:

    Create "tempkeys" table
    => select tempkey_install();

    Drop "tempkeys" table
    => select tempkey_uninstall(); -- Fix it: no work now!


Base operations:

    Check is param exists
    => select tempkey_exists('common','nds%');
    0
    Get _non-existing_ value _without_ default_value
    => select tempkey_get('common','nds%');

    Get _non-existing_ value _with_ default_value
    => select tempkey_get('common','nds%',18);
    18

    Set value
    => select tempkey_set('common','nds%',10);
    10

    Get _existing_ value
    => select tempkey_get('common','nds%');
    10

    Add _existing_ key
    => select tempkey_add('common','nds%',18);

    Replace _existing_ key
    => select tempkey_replace('common','nds%',20);
    20

    Delete existing key
    => select tempkey_delete('common','nds%');
    1

    Delete non-existing key
    => select tempkey_delete('common','nds%');

    Delete all keys in unit
    => select tempkey_delete('common');

    Replace _non-existing_ key
    => select tempkey_replace('common','nds%',20);

    Add _non-existing_ key
    => select tempkey_add('common','nds%',18);
    18


Numeric operations:

    Increment _non-existing_ value
    => select tempkey_increment('common','num',10);
    10

    Increment _existing_ value
    => select tempkey_increment('common','num',10);
    20


String operations:

    Prepend value
    => select tempkey_prepend('common','string','prepend-');
    prepend-

    Append value
    => select tempkey_append('common','string','-append');
    prepend--append


Note:

    tempkey_flush is synonym to tempkey_delete

*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_TEMPKEY)

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#ifndef SQLITE_CORE
#include "sqlite3ext.h"
SQLITE_EXTENSION_INIT1
#else
#include "sqlite3.h"
#endif

#define TEMPKEY_SCHEMA_CREATE \
"CREATE TEMP TABLE IF NOT EXISTS _tempkeys (  -- temp parameters storage \n" \
"        unit text not null,\n"                                             \
"        param text not null,\n"                                            \
"        value text not null,\n"                                            \
"        UNIQUE (unit,param) on conflict replace\n"                         \
");"

//#define TEMPKEY_SCHEMA_DROP "DROP TABLE IF EXISTS _tempkeys;"
#define TEMPKEY_SCHEMA_DROP "DELETE FROM _tempkeys;"

#define TEMPKEY_DELETE_UNIT      "delete from _tempkeys where unit=?"
#define TEMPKEY_DELETE           "delete from _tempkeys where unit=? and param=?"
#define TEMPKEY_INSERT           "insert into _tempkeys (unit,param,value) values (?,?,?)"
#define TEMPKEY_UPDATE           "update _tempkeys set value=? where unit=? and param=?"
#define TEMPKEY_INCREMENT        "update _tempkeys set value=value+? where unit=? and param=?"
#define TEMPKEY_APPEND           "update _tempkeys set value=value || ? where unit=? and param=?"
#define TEMPKEY_PREPEND          "update _tempkeys set value=? || value where unit=? and param=?"
#define TEMPKEY_COUNT            "select count(*) from _tempkeys where unit=? and param=?"
#define TEMPKEY_SELECT           "select value from _tempkeys where unit=? and param=?"

int tempkey_count(sqlite3_context *context, unsigned char* unit, unsigned char* param, int* count){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */
    *count = 0;

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_COUNT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    if(SQLITE_ROW == rc) {
        *count = sqlite3_column_int(pStmt,0);
    }
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int tempkey_insert(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_INSERT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, value, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int tempkey_update(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_UPDATE, -1, &pStmt, 0);
    if( SQLITE_OK != rc){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int tempkey_increment(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_INCREMENT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int tempkey_append(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_APPEND, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}


int tempkey_prepend(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char* value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_PREPEND, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, value, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 3, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

int tempkey_delete(sqlite3_context *context, unsigned char* unit, unsigned char* param){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_DELETE, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit, -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param,  -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}


int tempkey_delete_unit(sqlite3_context *context, unsigned char* unit){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_DELETE_UNIT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    sqlite3_finalize(pStmt);

    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

// returned value must be cleaned!
int tempkey_get(sqlite3_context *context, unsigned char* unit, unsigned char* param, unsigned char** value){
    sqlite3 *db;
    sqlite3_stmt *pStmt;        /* A statement */
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_prepare_v2(db, TEMPKEY_SELECT, -1, &pStmt, 0);
    if( SQLITE_OK != rc ){
        return rc;
    }
    sqlite3_bind_text(pStmt, 1, unit,  -1, SQLITE_STATIC);
    sqlite3_bind_text(pStmt, 2, param, -1, SQLITE_STATIC);

    rc = sqlite3_step(pStmt);
    if( SQLITE_ROW == rc ) {
        *value = (unsigned char*)sqlite3_malloc (strlen((unsigned char*)sqlite3_column_text(pStmt,0))+1);
        strcpy( *value, (unsigned char*)sqlite3_column_text(pStmt,0));
    }
    sqlite3_finalize(pStmt);
    if ( SQLITE_DONE == rc || SQLITE_ROW == rc ) {
        return SQLITE_OK;
    } else {
        return rc;
    }
}

// tempkey_exists (unit,param)
static void tempkey_existsFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int count;
    int rc;                     /* Result code */

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    sqlite3_result_int(context, count);
}

// tempkey_get (unit,param,default_value)
static void tempkey_getFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return default_value when defined 3-th argument
    if ( 0 == count  && 3 == argc ) {
        sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
        return;
    }
    // key not found
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }

    rc = tempkey_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    sqlite3_result_text( context, value, -1, sqlite3_free);

}

// store this data
// tempkey_set (unit,param,value)
static void tempkey_setFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = tempkey_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = tempkey_update(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // returns value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// increment this data
// tempkey_increment (unit,param,value)
static void tempkey_incrementFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = tempkey_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = tempkey_increment(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = tempkey_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// append this data
// tempkey_append (unit,param,value)
static void tempkey_appendFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value = 0;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = tempkey_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = tempkey_append(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = tempkey_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// prepend this data
// tempkey_prepend (unit,param,value)
static void tempkey_prependFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    unsigned char *value = 0;
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        rc = tempkey_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    } else {
        rc = tempkey_prepend(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    }
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    rc = tempkey_get(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), &value);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 != value) {
        sqlite3_result_text( context, value, -1, sqlite3_free);
        return;
    }
    // returns null value
    sqlite3_result_null(context);
    return;
}

// store this data, but only if the server *doesn't* already hold data for this key
// tempkey_add (unit,param,value)
static void tempkey_addFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server does already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (count > 0) {
        sqlite3_result_null(context);
        return;
    }
    rc = tempkey_insert(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return added value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// store this data, but only if the server *does* already hold data for this key
// tempkey_replace (unit,param,value)
static void tempkey_replaceFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) || SQLITE_NULL==sqlite3_value_type(argv[2]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't already hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }
    rc = tempkey_update(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]), (unsigned char*) sqlite3_value_text(argv[2]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return replaced value
    sqlite3_result_text( context, (char*)sqlite3_value_text(argv[2]), -1, SQLITE_TRANSIENT);
    return;
}


// delete this data
// tempkey_delete (unit,param)
static void tempkey_deleteFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */
    int count;

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) || SQLITE_NULL==sqlite3_value_type(argv[1]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    // server doesn't hold this key
    rc = tempkey_count(context, (unsigned char *)sqlite3_value_text(argv[0]), (unsigned char *)sqlite3_value_text(argv[1]), &count);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    if (0 == count) {
        sqlite3_result_null(context);
        return;
    }
    rc = tempkey_delete(context, (unsigned char*) sqlite3_value_text(argv[0]), (unsigned char*) sqlite3_value_text(argv[1]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return 1
    sqlite3_result_int(context, 1);
}

// delete this unit
// tempkey_delete (unit)
static void tempkey_delete_unitFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    if( SQLITE_NULL==sqlite3_value_type(argv[0]) ){
        sqlite3_result_null(context);
        return;
    }
    db = (sqlite3*) sqlite3_context_db_handle(context);

    rc = tempkey_delete_unit(context, (unsigned char*) sqlite3_value_text(argv[0]));
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

static void tempkey_installFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_exec(db, TEMPKEY_SCHEMA_CREATE, 0, 0, 0);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

static void tempkey_uninstallFunc(sqlite3_context *context, int argc, sqlite3_value **argv){
    sqlite3 *db;
    int rc;                     /* Result code */

    db = (sqlite3*) sqlite3_context_db_handle(context);
    rc = sqlite3_exec(db, TEMPKEY_SCHEMA_DROP, 0, 0, 0);
    if( SQLITE_OK != rc ){
        sqlite3_result_error(context, sqlite3_errmsg(db), -1);
        return;
    }
    // return null value
    sqlite3_result_null(context);
}

/* SQLite invokes this routine once when it loads the extension.
** Create new functions, collating sequences, and virtual table
** modules here.  This is usually the only exported symbol in
** the shared library.
*/

int sqlite3TempKeyInit(sqlite3 *db){
static const struct {
    char *zName;
    signed char nArg;
    int argType;           /* 1: 0, 2: 1, 3: 2,...  N:  N-1. */
    int eTextRep;          /* 1: UTF-16.  0: UTF-8 */
    void (*xFunc)(sqlite3_context*,int,sqlite3_value **);
} aFuncs[] = {
    { "tempkey_install",             0, 0, SQLITE_UTF8,    tempkey_installFunc },
    { "tempkey_uninstall",           0, 0, SQLITE_UTF8,    tempkey_uninstallFunc },
    { "tempkey_exists",              2, 0, SQLITE_UTF8,    tempkey_existsFunc },
    { "tempkey_get",                 2, 0, SQLITE_UTF8,    tempkey_getFunc },
    { "tempkey_get",                 3, 0, SQLITE_UTF8,    tempkey_getFunc },
    { "tempkey_delete",              2, 0, SQLITE_UTF8,    tempkey_deleteFunc },
    { "tempkey_delete",              1, 0, SQLITE_UTF8,    tempkey_delete_unitFunc },
    { "tempkey_flush",               2, 0, SQLITE_UTF8,    tempkey_deleteFunc },
    { "tempkey_flush",               1, 0, SQLITE_UTF8,    tempkey_delete_unitFunc },
    { "tempkey_set",                 3, 0, SQLITE_UTF8,    tempkey_setFunc },
    { "tempkey_increment",           3, 0, SQLITE_UTF8,    tempkey_incrementFunc },
    { "tempkey_append",              3, 0, SQLITE_UTF8,    tempkey_appendFunc },
    { "tempkey_prepend",             3, 0, SQLITE_UTF8,    tempkey_prependFunc },
    { "tempkey_add",                 3, 0, SQLITE_UTF8,    tempkey_addFunc },
    { "tempkey_replace",             3, 0, SQLITE_UTF8,    tempkey_replaceFunc },
};

int i;
for(i=0; i<sizeof(aFuncs)/sizeof(aFuncs[0]); i++){
    void *pArg;
    int argType = aFuncs[i].argType;
    pArg = (void*)(int)argType;
    sqlite3_create_function(db, aFuncs[i].zName, aFuncs[i].nArg,
        aFuncs[i].eTextRep, pArg, aFuncs[i].xFunc, 0, 0);
}

return 0;
}

#if !SQLITE_CORE
int sqlite3_extension_init(
sqlite3 *db,
char **pzErrMsg,
const sqlite3_api_routines *pApi
){
SQLITE_EXTENSION_INIT2(pApi)
return sqlite3TempKeyInit(db);
}
#endif

#endif
