/*
The extension is based on public-domain code from http://murmurhash.googlepages.com/

Maintained by Alexey Pechnikov (pechnikov@mobigroup.ru) and tested on linux only.
Thanks to Anton Kovalenko for help to incremental realisation on C. The code is public domain.

Compile as
     gcc -fPIC -lm -shared murmurhash.c -o libsqlitemurmurhash.so

Load into sqlite3 shell as
 .load './libsqlitemurmurhash.so'

Use as
 select murmurhash('testkey');
 => 3599487917
 select murmurhash('testkey1','testkey2');
 => 2144763366
 select murmurhash('testkey1','testkey2',..,'testkeyN');

*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_MURMURHASH)
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#ifndef SQLITE_CORE
  #include "sqlite3ext.h"
  SQLITE_EXTENSION_INIT1
#else
  #include "sqlite3.h"
#endif

#define mmix(h,k) { k *= murmurhash2a_m; k ^= k >> murmurhash2a_r; k *= murmurhash2a_m; h *= murmurhash2a_m; h ^= k; }

  struct MurmurHash2A {
    unsigned int hash;
    unsigned int tail;
    unsigned int count;
    unsigned int size;
  };

  static const unsigned int murmurhash2a_m = 0x5bd1e995;
  static const int murmurhash2a_r = 24;

void MurmurHash2A_MixTail (struct MurmurHash2A * me,
                           const unsigned char **data, int *len )
    {
        while( *len && ((*len<4) || me->count) )
        {
            me->tail |= (*(*data)++) << (me->count * 8);
            me->count++;
            (*len)--;
            if(me->count == 4)
            {
                mmix(me->hash,me->tail);
                me->tail = 0;
                me->count = 0;
            }
        }
    }

    void MurmurHash2A_Begin (struct MurmurHash2A * me)
    {
        me->hash  = 0;
        me->tail  = 0;
        me->count = 0;
        me->size  = 0;
    }

    void MurmurHash2A_Add (struct MurmurHash2A * me, const unsigned char * data, int len )
    {
        me->size += len;
        MurmurHash2A_MixTail(me,&data,&len);
        while(len >= 4)
        {
            unsigned int k = data[0]|(data[1]<<8)|(data[2]<<16)|(data[3]<<24);
            mmix(me->hash,k);
            data += 4;
            len -= 4;
        }
        MurmurHash2A_MixTail(me,&data,&len);
    }

    unsigned int MurmurHash2A_End (struct MurmurHash2A * me)
    {
        mmix(me->hash,me->tail);
        mmix(me->hash,me->size);

        me->hash ^= me->hash >> 13;
        me->hash *= murmurhash2a_m;
        me->hash ^= me->hash >> 15;

        return me->hash;
    }

static void murmurhashFunc(
 sqlite3_context *context,
 int argc,
 sqlite3_value **argv
) {
  int i;
  unsigned int pSize;
  unsigned int hash;
  struct MurmurHash2A ctx;

  if( argc<1 ) return;
  if( sqlite3_value_type(argv[0]) == SQLITE_NULL ){
    sqlite3_result_null(context);
    return;
  }

  MurmurHash2A_Begin(&ctx);
  pSize = argc;
  MurmurHash2A_Add(&ctx, (unsigned char*)&pSize, sizeof(pSize));
  for(i=0; i < argc; i++){
    const char *zData = (char*)sqlite3_value_blob(argv[i]);
    if( zData ){
        pSize = sqlite3_value_bytes(argv[i]);
        MurmurHash2A_Add(&ctx, (unsigned char*)&pSize, sizeof(pSize));
        MurmurHash2A_Add(&ctx, (unsigned char*)zData, sqlite3_value_bytes(argv[i]));
    }
  }
  hash = MurmurHash2A_End(&ctx);

  sqlite3_result_int64(context, hash);
}

static void murmurhashstep(sqlite3_context *context, int argc, sqlite3_value **argv){
  unsigned int pSize;
  unsigned int *gctx;
  struct MurmurHash2A ctx;
  int i;
  if( argc<1 ) return;
  gctx = sqlite3_aggregate_context(context, sizeof(*gctx));
  if( gctx==0 ) return;
  if( sqlite3_aggregate_count(context)==1 ){
    *gctx = 0;
  }

  MurmurHash2A_Begin(&ctx);
  pSize = argc;
  MurmurHash2A_Add(&ctx, (unsigned char*)&pSize, sizeof(pSize));
  for(i=0; i < argc; i++){
    const char *zData = (char*)sqlite3_value_blob(argv[i]);
    if( zData ){
        pSize = sqlite3_value_bytes(argv[i]);
        MurmurHash2A_Add(&ctx, (unsigned char*)&pSize, sizeof(pSize));
        MurmurHash2A_Add(&ctx, (unsigned char*)zData, sqlite3_value_bytes(argv[i]));
    }
  }
  *gctx = *gctx ^ MurmurHash2A_End(&ctx);
}
static void murmurhashfinalize(sqlite3_context *context){
  unsigned int hash;
  unsigned int *gctx;
  gctx = sqlite3_aggregate_context(context, sizeof(*gctx));
  sqlite3_result_int64(context, *gctx);
}

int sqlite3MurmurhashInit(sqlite3 *db){
  sqlite3_create_function(db, "group_murmurhash", -1, SQLITE_UTF8, 0, 0, murmurhashstep, murmurhashfinalize);
  sqlite3_create_function(db, "murmurhash",       -1, SQLITE_UTF8, 0, murmurhashFunc,  0, 0);
  return 0;
}

#if !SQLITE_CORE
int sqlite3_extension_init(
  sqlite3 *db,
  char **pzErrMsg,
  const sqlite3_api_routines *pApi
){
  SQLITE_EXTENSION_INIT2(pApi)
  return sqlite3MurmurhashInit(db);
}
#endif

#endif
