/*
 *      "Replication" SQLite extension
 *
 *      Copyright 2009 Alexey Pechnikov <pechnikov@mobigroup.ru>
 *      The code is public domain.
 */

/* 
Compile as
     gcc -shared versioning.c -o libsqliteversioning.so
*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_VERSIONING)

#include <assert.h>
#include <stdlib.h>

#ifndef NDEBUG
    #include <string.h>
#endif 

#ifndef SQLITE_CORE
    #include <sqlite3ext.h>
    SQLITE_EXTENSION_INIT1
#else
    #include "sqlite3.h"
#endif

static void versioning_create(sqlite3_context *context, int argc,
    sqlite3_value **argv)
{
    int rc;
    char *table;
    char *sql;
    sqlite3 *db;

    if (argc != 1) {
        sqlite3_result_error(context, "Function must get single argument. The argument is table name.", -1);
        return;
    }
    if ((sqlite3_value_type(argv[0]) != SQLITE_TEXT)) {
        sqlite3_result_error(context, "Table name must be a text string", -1);
        return;
    }

    table = (char*)sqlite3_value_text(argv[0]);

    db = sqlite3_context_db_handle(context);

    rc = sqlite3_exec(db, "SAVEPOINT versioning;", NULL, NULL, NULL);
    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    sql = sqlite3_mprintf(
        // add actions table
        "CREATE TABLE %s_versioning_action(date REAL, tbl_action TEXT, tbl_rowid INTEGER, history_rowid INTEGER, checksum TEXT, initiator TEXT);"
        "CREATE INDEX %s_versioning_action_tbl_rowid_date_idx on %s_versioning_action(tbl_rowid,date);"
        "CREATE INDEX %s_versioning_action_initiator_date_idx on %s_versioning_action(initiator,date);"
        "CREATE INDEX %s_versioning_action_checksum_idx       on %s_versioning_action(checksum);"
        // add history table
        "CREATE TABLE %s_versioning_history AS SELECT * FROM %s LIMIT 0;", // this work, really work :-)
        table, table, table, table, table, table, table, // create replication table and indices
        table, table // create history table
        );
//sqlite3_result_text( context, (char*)sql, -1, SQLITE_TRANSIENT);
//return;
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        sqlite3_exec(db, "ROLLBACK TO SAVEPOINT versioning;", NULL, NULL, NULL);
    }

    sqlite3_exec(db, "RELEASE SAVEPOINT versioning;", NULL, NULL, NULL);

    sqlite3_result_null(context);
}


static void versioning(sqlite3_context *context, int argc,
    sqlite3_value **argv)
{
    int rc, pk;
    char *func, *table, *initiator = "local";
    char *sql, *name, *names = NULL, *_names = NULL;
    sqlite3_stmt *stmt;
    sqlite3 *db;

    if (argc != 3 && argc != 2) {
        sqlite3_result_error(context, "Function must get two or three arguments. First argument is hash function name, "
            "second is table name and third is initiator name (by default is \"local\")", -1);
        return;
    }

    if ((sqlite3_value_type(argv[0]) != SQLITE_TEXT)) {
        sqlite3_result_error(context, "Hash function name must be a text string", -1);
        return;
    }

    if ((sqlite3_value_type(argv[1]) != SQLITE_TEXT)) {
        sqlite3_result_error(context, "Table name must be a text string", -1);
        return;
    }

    func  = (char*)sqlite3_value_text(argv[0]);
    table = (char*)sqlite3_value_text(argv[1]);
    if (argc > 2 && (sqlite3_value_type(argv[2]) != SQLITE_NULL)) {
        initiator = (char*)sqlite3_value_text(argv[2]);
    }

    db = sqlite3_context_db_handle(context);

    rc = sqlite3_exec(db, "SAVEPOINT versioning;", NULL, NULL, NULL);
    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    /*
    PRAGMA table_info([table]) returns:

    [0] = column number
    [1] = column name
    [2] = column type
    [3] = flag: is the column NOT NULL?
    [4] = default value of the column
    [5] = flag: is this column part of the table's PRIMARY KEY?
    */
    sql = sqlite3_mprintf("PRAGMA table_info(%s)", table);
    rc = sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    while ((rc = sqlite3_step(stmt)) != SQLITE_DONE) {
        if (rc == SQLITE_ROW) {
            pk = sqlite3_column_int(stmt, 5);
            name = (char*)sqlite3_column_text(stmt, 1);
            if (names) {
                _names = names;
                names = sqlite3_mprintf("%s,%s", _names, name);
                sqlite3_free(_names);
            }
            else {
                names = sqlite3_mprintf("%s", name);
            }
        }
        else {
            break;
        }
    }

    sqlite3_finalize(stmt);

    sql = sqlite3_mprintf(
        // drop triggers
        "DROP TRIGGER IF EXISTS %s_versioning_i;"
        "DROP TRIGGER IF EXISTS %s_versioning_u;"
        "DROP TRIGGER IF EXISTS %s_versioning_d;"
        // add triggers to source table
        "CREATE TRIGGER  %s_versioning_i AFTER INSERT ON %s BEGIN "
            "INSERT INTO %s_versioning_history SELECT * FROM %s WHERE rowid=new.rowid;"
            "INSERT INTO %s_versioning_action values (julianday('now'), 'I', new.rowid, last_insert_rowid(), (SELECT %s(%s) FROM %s WHERE rowid=new.rowid), '%s');"
        "END;"
        "CREATE TRIGGER  %s_versioning_u AFTER UPDATE ON %s BEGIN "
            "INSERT INTO %s_versioning_history SELECT * FROM %s WHERE rowid=new.rowid;"
            "INSERT INTO %s_versioning_action values (julianday('now'), 'U', new.rowid, last_insert_rowid(), (SELECT %s(%s) FROM %s WHERE rowid=new.rowid), '%s');"
        "END;"
        "CREATE TRIGGER  %s_versioning_d AFTER DELETE ON %s BEGIN "
            "INSERT INTO %s_versioning_action values (julianday('now'), 'D', old.rowid, NULL, (SELECT %s(%s) FROM %s WHERE rowid=old.rowid), '%s');"
        "END;",
        // drop triggers
        table, table, table,
        // insert trigger
        table, table, table, table, table, func, names, table, initiator,
        // update trigger
        table, table, table, table, table, func, names, table, initiator,
        // delete trigger
        table, table, table, func, names, table, initiator
        );
        sqlite3_free(names);

//sqlite3_result_text( context, (char*)sql, -1, SQLITE_TRANSIENT);
//return;
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        sqlite3_exec(db, "ROLLBACK TO SAVEPOINT versioning;", NULL, NULL, NULL);
    }

    sqlite3_exec(db, "RELEASE SAVEPOINT versioning;", NULL, NULL, NULL);

    sqlite3_result_null(context);
}

static void versioning_drop(sqlite3_context *context, int argc,
            sqlite3_value **argv)
{
    int rc;
    char *table;
    char *sql = NULL;
    sqlite3_stmt *stmt;
    sqlite3 *db;

    if (argc != 1) {
        sqlite3_result_error(context, "Function must get single argument. The argument is table name.", -1);
        return;
    }
    if ((sqlite3_value_type(argv[0]) != SQLITE_TEXT)) {
        sqlite3_result_error(context, "Table name must be a text string", -1);
        return;
    }

    table = (char*)sqlite3_value_text(argv[0]);

    db = sqlite3_context_db_handle(context);

    rc = sqlite3_exec(db, "SAVEPOINT versioning;", NULL, NULL, NULL);
    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        return;
    }

    sql = sqlite3_mprintf(
        // drop table can doesn't work with error message "SQL error: database table is locked". Ticket 4001.
        "DROP TABLE %s_versioning_action;"
        "DROP TABLE %s_versioning_history;",
        table, table
        );
//sqlite3_result_text( context, (char*)sql, -1, SQLITE_TRANSIENT);
//return;
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    sqlite3_free(sql);

    if (rc != SQLITE_OK) {
        sqlite3_result_error_code(context, rc);
        sqlite3_exec(db, "ROLLBACK TO SAVEPOINT versioning;", NULL, NULL, NULL);
        return;
    }

    sqlite3_exec(db, "RELEASE SAVEPOINT versioning;", NULL, NULL, NULL);

    sqlite3_result_null(context);
}

static void arg1(sqlite3_context *context, int argc,
    sqlite3_value **argv)
{
    if (argc >= 1) {
        sqlite3_result_text(context, (char*)sqlite3_value_text(argv[0]), -1, SQLITE_TRANSIENT);
        return;
    }
    sqlite3_result_null(context);
}

#if !SQLITE_CORE
static
#endif
int sqlite3VersioningInit(sqlite3 *db)
{
    int rc, i;
    static const struct {
        char *zName;
        signed char nArg;
        int argType;           /* 1: 0, 2: 1, 3: 2,...  N:  N-1. */
        int eTextRep;          /* 1: UTF-16.  0: UTF-8 */
        void (*xFunc)(sqlite3_context*,int,sqlite3_value **);
    } aFuncs[] = {
        { "versioning_create",        1, 0, SQLITE_UTF8,    versioning_create },
        { "versioning_drop",          1, 0, SQLITE_UTF8,    versioning_drop   },
        { "versioning",               2, 0, SQLITE_UTF8,    versioning        },
        { "versioning",               3, 0, SQLITE_UTF8,    versioning        },
        { "arg1",                    -1, 0, SQLITE_UTF8,    arg1              },
    };

    for(i=0; i<sizeof(aFuncs)/sizeof(aFuncs[0]); i++){
        void *pArg;
        int argType = aFuncs[i].argType;
        pArg = (void*)(int)argType;
        rc = sqlite3_create_function(db, aFuncs[i].zName, aFuncs[i].nArg,
            aFuncs[i].eTextRep, pArg, aFuncs[i].xFunc, 0, 0);
        if (rc != SQLITE_OK) {
            break;
        }
    }

    return rc;
}

#if !SQLITE_CORE
int sqlite3_extension_init(sqlite3 *db, char **pzErrMsg,
                        const sqlite3_api_routines *pApi)
{
    SQLITE_EXTENSION_INIT2(pApi)
    return sqlite3VersioningInit(db);
}
#endif

#endif
