/*
The functions coded by Alexey Pechnikov (pechnikov@mobigroup.ru) and tested on linux only.
The code is public domain.

Compile as
     gcc -fPIC -lm -shared env.c -o libsqliteenv.so

Load into sqlite3 shell as
    .load './libsqliteenv.so'

Use as
    select current_user();
    => veter

    select current_locale();
    => ru_RU.UTF-8

    select getenv('LANG');
    => ru_RU.UTF-8

TODO: HOST_NAME (MSSQL)
*/

#if !defined(SQLITE_CORE) || defined(SQLITE_ENABLE_ENV)

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <pwd.h>

#include <assert.h>
#include <locale.h>
#include <langinfo.h>

#include <time.h>

#ifndef SQLITE_CORE
  #include "sqlite3ext.h"
  SQLITE_EXTENSION_INIT1
#else
  #include "sqlite3.h"
#endif

static void getenvFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    char *s;

    if( sqlite3_value_type(argv[0]) == SQLITE_NULL ){
        sqlite3_result_null(context);
        return;
    }

    if (s=getenv((char*)sqlite3_value_text(argv[0])))
    {
        sqlite3_result_text( context, s, -1, SQLITE_TRANSIENT);
        return;
    }
    sqlite3_result_null(context);
}

/*
static void current_localeFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    char *s;

    if ((s = getenv("LC_ALL")) || (s = getenv("LC_CTYPE")) || (s = getenv ("LANG")))
    {
        sqlite3_result_text( context, s, -1, SQLITE_TRANSIENT);
        return;
    }
    sqlite3_result_null(context);
}
*/

static void current_localeFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    sqlite3_result_text( context, setlocale(LC_ALL, ""), -1, SQLITE_TRANSIENT);
}

static void current_charsetFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    setlocale(LC_ALL, "");
    sqlite3_result_text( context, nl_langinfo (CODESET), -1, SQLITE_TRANSIENT);
}

/*
Sybase current_user() function

select current_user();
veter
*/
static void current_userFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    char *s;

    if (s=getenv("USER"))
    {
        sqlite3_result_text( context, s, -1, SQLITE_TRANSIENT);
        return;
    }
    struct passwd *who;
    if ((who = getpwuid(getuid ())) != NULL)
    {
        sqlite3_result_text( context, (char*)who->pw_name, -1, SQLITE_TRANSIENT);
        return;
    }
    sqlite3_result_null(context);
}

static void current_databaseFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    char *s;

    if (s=getenv("DATABASE"))
    {
        sqlite3_result_text( context, s, -1, SQLITE_TRANSIENT);
        return;
    }
    sqlite3_result_null(context);
}

static void tzsetFunc(
    sqlite3_context *context,
    int argc,
    sqlite3_value **argv
) {
    tzset();
    sqlite3_result_null(context);
}


/* SQLite invokes this routine once when it loads the extension.
** Create new functions, collating sequences, and virtual table
** modules here.  This is usually the only exported symbol in
** the shared library.
*/

int sqlite3EnvInit(sqlite3 *db){
  static const struct {
     char *zName;
     signed char nArg;
     int argType;           /* 1: 0, 2: 1, 3: 2,...  N:  N-1. */
     int eTextRep;          /* 1: UTF-16.  0: UTF-8 */
     void (*xFunc)(sqlite3_context*,int,sqlite3_value **);
  } aFuncs[] = {
    { "current_locale",    0, 0, SQLITE_UTF8,    current_localeFunc      },
    { "current_charset",   0, 0, SQLITE_UTF8,    current_charsetFunc     },
    { "getenv",            1, 0, SQLITE_UTF8,    getenvFunc              },
    { "tzset",             0, 0, SQLITE_UTF8,    tzsetFunc               },
    // PostgreSQL, MySQL, Sybace, MSSQL, etc. functions
    { "schema",            0, 0, SQLITE_UTF8,    current_databaseFunc    },
    { "current_schema",    0, 0, SQLITE_UTF8,    current_databaseFunc    },
    { "database",          0, 0, SQLITE_UTF8,    current_databaseFunc    },
    { "current_database",  0, 0, SQLITE_UTF8,    current_databaseFunc    },
    { "current_catalog",   0, 0, SQLITE_UTF8,    current_databaseFunc    },
    { "user",              0, 0, SQLITE_UTF8,    current_userFunc        },
    { "current_user",      0, 0, SQLITE_UTF8,    current_userFunc        },
    { "session_user",      0, 0, SQLITE_UTF8,    current_userFunc        },
    { "system_user",       0, 0, SQLITE_UTF8,    current_userFunc        },
  };

  int i;
  for(i=0; i<sizeof(aFuncs)/sizeof(aFuncs[0]); i++){
    void *pArg;
    int argType = aFuncs[i].argType;
    pArg = (void*)(int)argType;
    sqlite3_create_function(db, aFuncs[i].zName, aFuncs[i].nArg,
        aFuncs[i].eTextRep, pArg, aFuncs[i].xFunc, 0, 0);
  }

  return 0;
}

#if !SQLITE_CORE
int sqlite3_extension_init(
  sqlite3 *db, 
  char **pzErrMsg,
  const sqlite3_api_routines *pApi
){
  SQLITE_EXTENSION_INIT2(pApi)
  return sqlite3EnvInit(db);
}
#endif

#endif
